import typing as T

import numpy as np

import terrapyn as tp


def celsius_to_kelvin(celsius):
    """
    Convert temperature in degrees Celsius to degrees Kelvin.

    Args:
        celsius: Degrees Celsius

    Returns:
        Degrees Kelvin
    """
    return celsius + 273.15


def kelvin_to_celsius(kelvin):
    """
    Convert temperature in degrees Kelvin to degrees Celsius.

    Args:
        kelvin: Degrees Kelvin

    Returns:
        Degrees Celsius
    """
    return kelvin - 273.15


def degrees_to_radians(degrees):
    """
    Convert angular degrees to radians

    Args:
        degrees: Value in degrees to be converted.

    Returns:
        Value in radians
    """
    return degrees * (np.pi / 180.0)


def radians_to_degrees(radians):
    """
    Convert radians to angular degrees

    Args:
        radians: Value in radians to be converted.

    Returns:
        Value in angular degrees
    """
    return radians * (180.0 / np.pi)


def kilometers_per_hour_to_meters_per_second(speed):
    """
    Convert km/h to m/s.

    Args:
        speed: Speed in km/h

    Returns:
        Speed in m/s
    """
    return speed / 3.6


def wind_speed_2m(ws, z):
    """
    Convert wind speed measured at different heights above the soil surface to wind speed at 2 m above the surface,
    assuming a short grass surface. Based on FAO equation 47 in Allen et al (1998).

    Args:
        ws: Measured wind speed [m s-1]
        z: Height of wind measurement above ground surface [m]

    Returns:
        Wind speed at 2 m above the surface [m s-1]
    """
    if ws is None:
        raise ValueError("`ws` must be provided")
    if z is None:
        raise ValueError("`z` must be provided")

    tp.validation.check_positive(ws, string="wind speed")
    tp.validation.check_positive(z, string="height of wind above surface")
    return (ws * 4.87) / np.log((67.8 * z) - 5.42)


def saturation_vapour_pressure_from_temperature(t, version=1):
    """
    Estimate saturation vapour pressure (**es**) from air temperature.

    `version=1` based on Alduchov & Eskridge 1996: Improved Magnus' form approximation of
    saturation vapor pressure. J. Appl. Meteor., 35, 601–609.

    `version=2` is based on equations 11 and 12 in Allen et al (1998).

    Args:
        t: Temperature [deg C]
        version: Version of approximation to use.

    Returns:
        Saturation vapour pressure [kPa]
    """
    if t is None:
        raise ValueError("`t` must be provided")

    if version == 1:
        return 0.61094 * np.exp((17.625 * t) / (t + 243.04))
    elif version == 2:
        return 0.6108 * np.exp((17.27 * t) / (t + 237.3))
    else:
        raise ValueError("`version` must be 1 or 2")


def delta_saturation_vapour_pressure(t):
    """
    Estimate the slope of the saturation vapour pressure curve at a given temperature.
    Based on equation 13 in Allen et al (1998). If using in the Penman-Monteith *t* should be the
    mean air temperature.

    Args:
        t: Air temperature [deg C]. Use mean air temperature for use in Penman-Monteith.

    Returns:
        Saturation vapour pressure [kPa degC-1]
    """
    if t is None:
        raise ValueError("`t` must be provided")

    numerator = 4098 * saturation_vapour_pressure_from_temperature(t)
    demoninator = np.power((t + 237.3), 2)
    return numerator / demoninator


def relative_humidity_from_actual_vapour_pressure_and_saturated_vapour_pressure(avp, svp):
    """
    Calculate relative humidity from the ratio of actual vapour pressure to saturation vapour pressure
    at the same temperature. See Allen et al (1998), page 67 for details.

    Args:
        avp: Actual vapour pressure [kPa]
        svp: Saturated vapour pressure [kPa]

    Returns:
        Relative humidity [%]
    """
    return 100.0 * avp / svp


def relative_humidity_from_dew_point_and_temperature(tdew, t):
    """
    Calculate relative humidity from the dew point temperature and temperature.
    Converts the temperatures to vapour pressures using `saturation_vapour_pressure_from_temperature`.

    Args:
        tdew: Dew point temperature [deg C]
        t: Air temperature [deg C]

    Returns:
        Relative humidity [%]
    """
    avp = saturation_vapour_pressure_from_temperature(tdew)
    svp = saturation_vapour_pressure_from_temperature(t)
    return relative_humidity_from_actual_vapour_pressure_and_saturated_vapour_pressure(avp, svp)


def psychrometric_constant_of_psychrometer(psychrometer=None, atmos_pres=None):
    """
    Calculate the psychrometric constant for different types of psychrometer at a given atmospheric pressure.
    Based on FAO equation 16 in Allen et al (1998).

    Args:
        psychrometer: Integer between 1 and 3 which denotes type of psychrometer:

            1. ventilated (Asmann or aspirated type) psychrometer with an air movement of approximately 5 m/s
            2. natural ventilated psychrometer with an air movement of approximately 1 m/s
            3. non ventilated psychrometer installed indoors

        atmos_pres: Atmospheric pressure [kPa]. Can be estimated using ``atmospheric_pressure()``.

    Returns:
        Psychrometric constant [kPa degC-1].
    """
    # Select coefficient based on type of ventilation of the wet bulb
    if psychrometer == 1:
        psy_coeff = 0.000662
    elif psychrometer == 2:
        psy_coeff = 0.000800
    elif psychrometer == 3:
        psy_coeff = 0.001200
    else:
        raise ValueError("psychrometer should be in range 1 to 3: {0}".format(psychrometer))

    if atmos_pres is None:
        raise ValueError("`atmos_pres` must be provided")

    return psy_coeff * atmos_pres


def psychrometric_constant(atmos_pres=None, altitude=None):
    """
    Calculate the psychrometric constant. This method assumes that the air is saturated with water vapour at the
    minimum daily temperature. This assumption may not hold in arid areas.
    Based on equation 8, page 95 in Allen et al (1998).

    Args:
        atmos_pres: Atmospheric pressure [kPa]. Not required if `altitude` is provided.
        altitude: Elevation/altitude above sea level [m]. Not required if `atmos_pres` is provided.

    Returns:
        Psychrometric constant [kPa degC-1].
    """
    if atmos_pres is None:
        if altitude is None:
            raise ValueError("`atmos_pres` or `altitude` must be provided")
        else:
            atmos_pres = atmospheric_pressure(altitude)
    tp.validation.check_positive(atmos_pres, "atmospheric pressure")
    return 0.000665 * atmos_pres


def atmospheric_pressure(altitude):
    """
    Estimate atmospheric pressure from altitude. Calculated using a simplification of the ideal gas law,
    assuming 20 degrees Celsius for a standard atmosphere. Based on equation 7, page 62 in Allen et al (1998).

    Args:
        altitude: Elevation/altitude above sea level [m]

    Returns:
        atmospheric pressure [kPa]
    """
    if altitude is None:
        raise ValueError("`altitude` must be provided")
    tmp = (293.0 - (0.0065 * altitude)) / 293.0
    return np.power(tmp, 5.26) * 101.3


def mean_saturation_vapour_pressure(tmin, tmax):
    """
    Estimate mean saturation vapour pressure, *es* [kPa] from minimum and maximum temperature.
    Based on equations 11 and 12 in Allen et al (1998). Mean saturation vapour pressure is
    calculated as the mean of the saturation vapour pressure at tmax (maximum temperature)
    and tmin (minimum temperature).

    Args:
        tmin: Minimum temperature [deg C]
        tmax: Maximum temperature [deg C]

    Returns:
        Mean saturation vapour pressure (*es*) [kPa]
    """
    if tmax is None:
        raise ValueError("`tmax` must be provided")
    if tmin is None:
        raise ValueError("`tmin` must be provided")

    return (saturation_vapour_pressure_from_temperature(tmin) + saturation_vapour_pressure_from_temperature(tmax)) / 2.0


def actual_vapour_pressure(
    tdew=None,
    twet=None,
    tdry=None,
    atmos_pres=None,
    psychrometer=None,
    tmin=None,
    tmax=None,
    rh_min=None,
    rh_max=None,
    rh_mean=None,
):
    """
    Estimate actual vapour pressure [kPa] (*ea*) from multiple methods, in order of preference, based on the
    provided input parameters:

    1 - Based on `tdew` - If dewpoint temperature data are available.
    2 - Based on `twet`, `tdry`, `atmos_pres` - If dry and wet bulb temperatures are available from a psychrometer.
    3 - Based on `tmin`, `tmax`, `rh_min`, `rh_max` - If  minimum and maximum relative humidity data available.
    4 - Based on `tmin`, `rh_max` - If measurement errors of relative humidity are large then use only
    maximum relative humidity    .
    5 - Based on `tmin`, `tmax`, `rh_mean` - If minimum and maximum relative humidity are not available but mean
    relative humidity is available (less reliable than options 3 or 4).
    6 - Based on `tmin` - If no data for the above. This function is less reliable in arid areas where it is
    recommended that 2 degrees Celsius is subtracted from the minimum temperature before it is passed to the
    function (following advice given in Annex 6 of Allen et al (1998).

    Args:
        tdew: Dewpoint temperature [deg C].
        twet: Wet bulb temperature [deg C].
        tdry: Dry bulb temperature [deg C].
        atmos_pres: Atmospheric pressure [kPa]
        psychrometer: Integer between 1 and 3 which denotes type of psychrometer -
        see `psychrometric_constant_of_psychrometer()`.
        tmin: Daily minimum temperature [deg C]
        tmax: Daily maximum temperature [deg C].
        rh_min: Minimum relative humidity [%]
        rh_max: Maximum relative humidity [%]
        rh_mean: Mean relative humidity [%] (mean of RH min and RH max).

    Returns:
        Actual vapour pressure [kPa]
    """
    if tdew is not None:
        return actual_vapour_pressure_from_tdew(tdew)
    elif all(i is not None for i in [twet, tdry, atmos_pres]):
        return actual_vapour_pressure_from_twet_tdry(twet, tdry, atmos_pres=atmos_pres, psychrometer=psychrometer)
    elif all(i is not None for i in [tmin, tmax, rh_min, rh_max]):
        return actual_vapour_pressure_from_rhmin_rhmax(rh_min=rh_min, rh_max=rh_max, tmin=tmin, tmax=tmax)
    elif all(i is not None for i in [tmin, rh_max]):
        return actual_vapour_pressure_from_rhmax(rh_max=rh_max, tmin=tmin)
    elif all(i is not None for i in [tmin, tmax, rh_mean]):
        return actual_vapour_pressure_from_rhmean(rh_mean=rh_mean, tmin=tmin, tmax=tmax)
    elif tmin is not None:
        return actual_vapour_pressure_from_tmin(tmin)
    else:
        raise ValueError("at least `tmin` must be provided")


def actual_vapour_pressure_from_tmin(tmin):
    """
    Estimate actual vapour pressure (*ea*) from minimum temperature. This method is to be used where humidity data
    are lacking or are of questionable quality. The method assumes that the dewpoint temperature is approximately
    equal to the minimum temperature (*tmin*), i.e. the air is saturated with water vapour at *tmin*.
    **Note**: This assumption may not hold in arid/semi-arid areas.
    In these areas it may be better to subtract 2 deg C from the minimum temperature (see Annex 6 in FAO paper).
    Based on equation 48 in Allen et al (1998).

    Args:
        tmin: Daily minimum temperature [deg C]

    Returns:
        Actual vapour pressure [kPa]
    """
    if tmin is None:
        raise ValueError("`tmin` must be provided")

    return saturation_vapour_pressure_from_temperature(tmin)


def actual_vapour_pressure_from_rhmin_rhmax(
    svp_tmin=None, svp_tmax=None, rh_min=None, rh_max=None, tmin=None, tmax=None
):
    """
    Estimate actual vapour pressure (*ea*) from saturation vapour pressure and relative humidity.
    For periods of a week, ten days or a month, `rh_max` and `rh_min` are obtained by dividing
    the sum of the daily values by the number of days in that period.
    Based on FAO equation 17 in Allen et al (1998).

    Args:
        svp_tmin: Saturation vapour pressure at daily minimum temperature [kPa]. Not required if `tmin` is provided.
        svp_tmax: Saturation vapour pressure at daily maximum temperature [kPa]. Not required if `tmax` is provided.
        rh_min: Minimum relative humidity [%]
        rh_max: Maximum relative humidity [%]
        tmin: Daily minimum temperature [deg C]. Required if `svp_tmin` is not provided.
        tmax: Daily maximum temperature [deg C]. Required if `svp_tmax` is not provided.

    Returns:
        Actual vapour pressure [kPa]
    """
    # Use `svp` if provided, otherwise use `tmin` and `tmax` to calculate `svp`
    if svp_tmin is None:
        if tmin is None:
            raise ValueError("`svp_tmin` or `tmin` must be provided")
        else:
            svp_tmin = saturation_vapour_pressure_from_temperature(tmin)
    if svp_tmax is None:
        if tmax is None:
            raise ValueError("`svp_tmax` or `tmax` must be provided")
        else:
            svp_tmax = saturation_vapour_pressure_from_temperature(tmax)
    tmp1 = svp_tmin * rh_max / 100.0
    tmp2 = svp_tmax * rh_min / 100.0
    return (tmp1 + tmp2) / 2.0


def actual_vapour_pressure_from_rhmax(svp_tmin=None, rh_max=None, tmin=None):
    """
    Estimate actual vapour pressure (*ea*) from saturation vapour pressure at daily minimum temperature
    and maximum relative humidity. Based on FAO equation 18 in Allen et al (1998).

    Args:
        svp_tmin: Saturation vapour pressure at daily minimum temperature [kPa]. Not required if `tmin` is provided.
        rh_max: Maximum relative humidity [%]
        tmin: Daily minimum temperature [deg C]. Required if `svp_tmin` is not provided.

    Returns:
        Actual vapour pressure [kPa]
    """
    if svp_tmin is None:
        if tmin is None:
            raise ValueError("`svp_tmin` or `tmin` must be provided")
        else:
            svp_tmin = saturation_vapour_pressure_from_temperature(tmin)

    return svp_tmin * rh_max / 100.0


def actual_vapour_pressure_from_rhmean(svp_tmin=None, svp_tmax=None, rh_mean=None, tmin=None, tmax=None):
    """
    Estimate actual vapour pressure (*ea*) from saturation vapour pressure at daily minimum and maximum
    temperature, and mean relative humidity. Based on FAO equation 19 in Allen et al (1998).

    Args:
        svp_tmin: Saturation vapour pressure at daily minimum temperature [kPa]. Not required if `tmin` is provided.
        svp_tmax: Saturation vapour pressure at daily maximum temperature [kPa]. Not required if `tmax` is provided.
        rh_mean: Mean relative humidity [%] (average of RH min and RH max).
        tmin: Daily minimum temperature [deg C]. Required if `svp_tmin` is not provided.
        tmax: Daily maximum temperature [deg C]. Required if `svp_tmax` is not provided.

    Returns:
        Actual vapour pressure [kPa]
    """
    # Use `svp` if provided, otherwise use `tmin` and `tmax` to calculate `svp`
    if svp_tmin is None:
        if tmin is None:
            raise ValueError("`svp_tmin` or `tmin` must be provided")
        else:
            svp_tmin = saturation_vapour_pressure_from_temperature(tmin)
    if svp_tmax is None:
        if tmax is None:
            raise ValueError("`svp_tmax` or `tmax` must be provided")
        else:
            svp_tmax = saturation_vapour_pressure_from_temperature(tmax)
    return (rh_mean / 100.0) * ((svp_tmax + svp_tmin) / 2.0)


def actual_vapour_pressure_from_tdew(tdew):
    """
    Estimate actual vapour pressure (*ea*) from dewpoint temperature.
    Based on equation 14 in Allen et al (1998). As the dewpoint temperature is the temperature to which air
    needs to be cooled to make it saturated, the actual vapour pressure is the saturation vapour pressure at
    the dewpoint temperature. This method is preferable to calculating vapour pressure from minimum temperature.

    Args:
        tdew: Dewpoint temperature [deg C]

    Returns:
        Actual vapour pressure [kPa]
    """
    if tdew is None:
        raise ValueError("`tdew` must be provided")

    return saturation_vapour_pressure_from_temperature(tdew)


def actual_vapour_pressure_from_twet_tdry(
    twet, tdry, svp_twet=None, psy_const=None, psychrometer=None, atmos_pres=None
):
    """
    Estimate actual vapour pressure (*ea*) from wet and dry bulb temperature.
    Based on equation 15 in Allen et al (1998). As the dewpoint temperature is the temperature to which air needs
    to be cooled to make it saturated, the actual vapour pressure is the saturation vapour pressure at the dewpoint
    temperature. This method is preferable to calculating vapour pressure from minimum temperature.
    Values for the psychrometric constant of the psychrometer ``psy_const`` can be calculated using
    ``psyc_const_of_psychrometer()``.

    Args:
        twet: Wet bulb temperature [deg C]
        tdry: Dry bulb temperature [deg C]
        svp_twet: Saturated vapour pressure at the wet bulb temperature [kPa]. Not required if `twet` is provided.
        psy_const: Psychrometric constant of the pyschrometer [kPa deg C-1].  Not required if both `atmos_pres` and
        `psychrometer` or just `atmos_pres` are provided.
        psychrometer: Integer between 1 and 3 which denotes type of psychrometer
        - see `psychrometric_constant_of_psychrometer()`
        atmos_pres: Atmospheric pressure [kPa]

    Returns:
        Actual vapour pressure [kPa]
    """
    if twet is None:
        raise ValueError("`twet` must be provided")
    if tdry is None:
        raise ValueError("`tdry` must be provided")

    if svp_twet is None:
        svp_twet = saturation_vapour_pressure_from_temperature(twet)

    if psy_const is None:
        if atmos_pres is None:
            raise ValueError("`atmos_pres` must be provided")
        if psychrometer is None:
            psy_const = psychrometric_constant(atmos_pres)
        else:
            psy_const = psychrometric_constant_of_psychrometer(psychrometer, atmos_pres)
    return svp_twet - psy_const * (tdry - twet)


def energy_to_evaporation(energy):
    """
    Convert energy (e.g. radiation energy) in MJ m-2 day-1 to the equivalent evaporation, assuming
    a grass reference crop. Energy is converted to equivalent evaporation using a conversion factor
    equal to the inverse of the latent heat of vapourisation (1 / lambda = 0.408).
    Based on FAO equation 20 in Allen et al (1998).

    Args:
        energy: Energy e.g. radiation or heat flux [MJ m-2 day-1].

    Returns:
        Equivalent evaporation [mm day-1].
    """
    if energy is None:
        raise ValueError("`energy` must be provided")

    return 0.408 * energy


def monthly_soil_heat_flux(t_month_prev, t_month, next_month=False):
    """
    Estimate monthly soil heat flux (Gmonth) from the mean air temperature of the previous and current or
    next month, assuming a grass crop.
    When `next_month==True`, `t_month` is for the next month, based on equation 43 in Allen et al (1998).
    WHen `next_month==False` `t_month` is for the current month, based on equation 44 in Allen et al (1998).
    The resulting heat flux can be converted to equivalent evaporation [mm day-1] using ``energy2evap()``.

    Args:
        t_month_prev: Mean air temperature of the previous month [deg Celsius]
        t_month: Mean air temperature of the current/next month [deg Celsius]
        next_month: If `True` then `t_month` is assumed to be the mean temperature of the next month,
        otherwise `t_month` is assumed to be the mean for the current month.

    Returns:
        Monthly soil heat flux (Gmonth) [MJ m-2 day-1]
    """
    if next_month:
        factor = 0.07
    else:
        factor = 0.14
    return factor * (t_month - t_month_prev)


def vectors_to_scalar(u, v):
    """Compute the scalar (length) from u and v vector components"""
    return np.sqrt(u * u + v * v)


def normalize_angle(angle_rad):
    """
    Return the input angle values in the range [0, 2 * pi].

    Args:
        angle_rad: Input angle in radians.

    Returns:
        Angle values in the range [0, 2 * pi]
    """
    return angle_rad % (2 * np.pi)


def vectors_to_angle(u, v):
    """Return the angle between two vectors in radians"""
    return np.arctan2(v, u)


def angle_to_bearing(angle, convention: str = "from", unit: str = "deg"):
    """
    Convert angle in radians to compass bearing convention, where the direction can be 'from'
    (the meteorological convention) or 'to' (the oceanographic convention).
    If angle is zero, bearing is set to 0 by convention.

    Args:
        angle: Angle values in radians to be converted from trigonometric to compass bearing convention.
        convention: Convention to return direction where 'from' (default) returns the direction the wind is coming
        from (meteorological convention), and 'to' returns the direction the wind is going towards
        (oceanographic convention).
        unit: Unit of angle, either 'rad' or 'deg' (default).

    Returns:
        Angle values in the interval [0, 360) degrees or [0, 2 * pi) radians.
    """
    if unit not in ("deg", "rad"):
        raise ValueError("Unit must be 'deg' or 'rad'.")

    if convention == "from":
        compass = -np.pi / 2 - angle
    elif convention == "to":
        compass = np.pi / 2 - angle
    else:
        raise ValueError("Convention must be 'from' or 'to'.")

    compass = normalize_angle(compass)

    if unit == "deg":
        return radians_to_degrees(compass)
    return compass


def wind_speed(u, v):
    """
    Compute the wind speed (scalar) from U and V vector components.

    Args:
        u: U component of the wind based on the [ECMWF](https://apps.ecmwf.int/codes/grib/param-db/?id=131)
            definition.
        v: V component of wind based on the [ECMWF](https://apps.ecmwf.int/codes/grib/param-db?id=132)
            definition.
    Returns:
        Wind speed in the same unit as u and v.
    """
    return vectors_to_scalar(u, v)


def wind_direction(u, v, convention: str = "from", unit: str = "deg"):
    """
    Compute the wind direction from U and V components. If U=V=0, wind direction is set to 0 by convention.

    Args:
        u: U component of the wind based on the [ECMWF](https://apps.ecmwf.int/codes/grib/param-db/?id=131)
            definition.
        v: V component of wind based on the [ECMWF](https://apps.ecmwf.int/codes/grib/param-db?id=132)
            definition.
        convention: Convention to return direction where 'from' returns the direction the wind is coming
        from (meteorological convention), and 'to' returns the direction the wind is going towards
        (oceanographic convention).
        unit: Unit of wind direction, 'rad' or 'deg'.
    Returns:
        The direction of the wind in the interval [0, 360) degrees or [0, 2*pi) radians.
    """
    zero_wind_mask = (np.abs(u) == 0) & (np.abs(v) == 0)
    angle = vectors_to_angle(u, v)
    bearing = angle_to_bearing(angle, convention=convention, unit=unit)
    bearing = np.where(zero_wind_mask, 0, bearing)
    return bearing


def cartesian_to_polar(x, y, compass: bool = False) -> T.Tuple:
    """
    Transform cartesian coordinates into polar coordinates (modulus and angle). Polar coordinates
    are given in the range [0, 2 * pi] using the trigonometric or bearing convention depending on
    **compass**.

    Args:
        x: Input values for x cartesian component.
        y: Input values for y cartesian component.
        compass: If True, angle values are given in bearing convention. If False (default), angle
        values are given in trigonometric convention.

    Returns:
        Tuple of modulus and angle values of the vector field given by **x** and **y**.
    """
    modulus = np.sqrt(np.square(x) + np.square(y))

    angle = vectors_to_angle(y, x)

    if compass:
        return modulus, angle_to_bearing(angle)
    else:
        return modulus, normalize_angle(angle)



import unittest

import numpy as np
import pandas as pd
import xarray as xr

import terrapyn as tp

SSRD = 13.28  # MJ m-2 day-1
WS = 2.078  # m s-1
T = 16.9  # degC
SVP = 1.997  # kPa
AVP = 1.409  # kPa
DELTA_SVP = 0.122  # kPa degC-1
PSY = 0.0666  # kPa degC-1
ELEVATION = 100  # m
ATMOS_PRES = 100.1  # kPa
TMAX = 21.5  # degC
TMIN = 12.3  # degC
WIND10M = 2.778  # m s-1
RHMAX = 84  # %
RHMIN = 63  # %
RHMEAN = 73.5  # %
SVP_TMIN = 1.431  # kPa
SVP_TMAX = 2.564  # kPa
LATITUDE = 50.8  # degrees
TWET = 19.5  # degC
TDRY = 25.6  # degC
TDEW = 14.8  # degC

n_lat = 3
n_lon = 3
n_time = 3
n_total = n_lat * n_lon * n_time
lats = pd.Index(np.linspace(43, 45, n_lat), name="lat")
lons = pd.Index(np.linspace(1, 3, n_lon), name="lon")
times = pd.date_range("20210401", freq="D", periods=n_time, name="time")
index = pd.MultiIndex.from_product([lats, lons, times])
df = pd.DataFrame(
    {
        "elevation": [ELEVATION] * n_total,  # m
        "tmax": [TMAX] * n_total,  # ˚C
        "tmin": [TMIN] * n_total,  # ˚C
        "rhmax": [RHMAX] * n_total,  # %
        "rhmin": [RHMIN] * n_total,  # %
        "wind10m": [WIND10M] * n_total,  # m/s
        "ssrd": [SSRD] * n_total,  # MJ m-2 day-1
        "ws": [WS] * n_total,  # m/s
        "t": [T] * n_total,  # degC
        "svp": [SVP] * n_total,  # kPa
        "avp": [AVP] * n_total,  # kPa
        "delta_svp": [DELTA_SVP] * n_total,  # kPa degC-1
        "psy": [PSY] * n_total,  # kPa degC-1
    },
    index=index,
)


class TestWindSpeed2m(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.wind_speed_2m(3.2, 10)
        np.testing.assert_almost_equal(result, 2.4, decimal=1)

    def test_series(self):
        result = tp.conversion.wind_speed_2m(df["wind10m"], 10).loc[44.0, 2.0, "2021-04-02"]
        np.testing.assert_almost_equal(result, 2.0778080868165487)


class TestSvpFromT(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.saturation_vapour_pressure_from_temperature(T)
        np.testing.assert_almost_equal(result, 1.9215514597960852)

    def test_series(self):
        result = tp.conversion.saturation_vapour_pressure_from_temperature(df["t"]).loc[44.0, 2.0, "2021-04-02"]
        np.testing.assert_almost_equal(result, 1.9215514597960852)


class TestDeltaSvp(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.delta_saturation_vapour_pressure(T)
        np.testing.assert_almost_equal(result, 0.12186328504483228)

    def test_series(self):
        result = tp.conversion.delta_saturation_vapour_pressure(df["t"]).loc[44.0, 2.0, "2021-04-02"]
        np.testing.assert_almost_equal(result, 0.12186328504483228)


class TestAtmPressure(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.atmospheric_pressure(ELEVATION)
        np.testing.assert_almost_equal(result, 100.12350828341812)

    def test_series(self):
        result = tp.conversion.atmospheric_pressure(df["elevation"]).loc[44.0, 2.0, "2021-04-02"]
        np.testing.assert_almost_equal(result, 100.12350828341812)


class TestPsyConstant(unittest.TestCase):
    def test_altitude_single_value(self):
        result = tp.conversion.psychrometric_constant(altitude=ELEVATION)
        np.testing.assert_almost_equal(result, 0.06658213300847304)

    def test_atmos_pres_single_value(self):
        result = tp.conversion.psychrometric_constant(atmos_pres=ATMOS_PRES)
        np.testing.assert_almost_equal(result, 0.0665665)

    def test_series_altitude(self):
        result = tp.conversion.psychrometric_constant(altitude=df["elevation"]).loc[44.0, 2.0, "2021-04-02"]
        np.testing.assert_almost_equal(result, 0.06658213300847304)


class TestAvpFromRhminRhmax(unittest.TestCase):
    def test_temp_and_rh_single_value(self):
        result = tp.conversion.actual_vapour_pressure_from_rhmin_rhmax(tmin=TMIN, tmax=TMAX, rh_min=RHMIN, rh_max=RHMAX)
        np.testing.assert_almost_equal(result, 1.4058885969470276)

    def test_svp_and_rh_single_value(self):
        result = tp.conversion.actual_vapour_pressure_from_rhmin_rhmax(
            svp_tmin=SVP_TMIN, svp_tmax=SVP_TMAX, rh_min=RHMIN, rh_max=RHMAX
        )
        np.testing.assert_almost_equal(result, 1.40868)


class TestAvpFromRhmax(unittest.TestCase):
    def test_svptmin_and_rhmax_single_value(self):
        result = tp.conversion.actual_vapour_pressure_from_rhmax(SVP_TMIN, RHMAX)
        np.testing.assert_almost_equal(result, 1.20204)

    def test_rhmax_and_tmin_single_value(self):
        result = tp.conversion.actual_vapour_pressure_from_rhmax(rh_max=RHMAX, tmin=TMIN)
        np.testing.assert_almost_equal(result, 1.1995003886672477)


class TestAvpFromTdew(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.actual_vapour_pressure_from_tdew(T)
        np.testing.assert_almost_equal(result, 1.9215514597960852)


class TestAvpFromTmin(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.actual_vapour_pressure_from_tmin(TMIN)
        np.testing.assert_almost_equal(result, 1.4279766531752949)


class TestAvpFromRhmean(unittest.TestCase):
    def test_svp_and_rh_single_value(self):
        result = tp.conversion.actual_vapour_pressure_from_rhmean(SVP_TMIN, SVP_TMAX, RHMEAN)
        np.testing.assert_almost_equal(result, 1.4681625)

    def test_rh_and_temp_single_value(self):
        result = tp.conversion.actual_vapour_pressure_from_rhmean(None, None, RHMEAN, TMIN, TMAX)
        np.testing.assert_almost_equal(result, 1.4652762230908918)


class TestDeg2Rad(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.degrees_to_radians(LATITUDE)
        np.testing.assert_almost_equal(result, 0.8866272600131193)


class TestRad2Deg(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.radians_to_degrees(0.3)
        np.testing.assert_almost_equal(result, 17.188733853924695)


class TestCelsius2Kelvin(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.celsius_to_kelvin(20)
        self.assertEqual(result, 293.15)


class TestKelvin2Celsius(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.kelvin_to_celsius(293.15)
        self.assertEqual(result, 20)


class TestKilometersPerHour2MetersPerSecond(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.kilometers_per_hour_to_meters_per_second(37)
        np.testing.assert_almost_equal(result, 10.277777777777777)


class TestAvpFromTwetTdry(unittest.TestCase):
    def test_single_value(self):
        atmos_pres = tp.conversion.atmospheric_pressure(1200)
        psy_const = tp.conversion.psychrometric_constant_of_psychrometer(1, atmos_pres)
        svp_twet = tp.conversion.saturation_vapour_pressure_from_temperature(TWET)
        result = tp.conversion.actual_vapour_pressure_from_twet_tdry(TWET, TDRY, svp_twet, psy_const)
        np.testing.assert_almost_equal(result, 1.9072375322669735)


class TestPsyConstOfPsychronometer(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.psychrometric_constant_of_psychrometer(1, ATMOS_PRES)
        np.testing.assert_almost_equal(result, 0.0662662)
        result = tp.conversion.psychrometric_constant_of_psychrometer(2, ATMOS_PRES)
        np.testing.assert_almost_equal(result, 0.08008)
        result = tp.conversion.psychrometric_constant_of_psychrometer(3, ATMOS_PRES)
        np.testing.assert_almost_equal(result, 0.12011999999999998)


class TestActualVapourPressure(unittest.TestCase):
    def test_first_method_single_value(self):
        result = tp.conversion.actual_vapour_pressure(tdew=TDEW)
        np.testing.assert_almost_equal(result, 1.6802076416421783)

    def test_second_method_single_value(self):
        result = tp.conversion.actual_vapour_pressure(twet=TWET, tdry=TDRY, atmos_pres=ATMOS_PRES)
        np.testing.assert_almost_equal(result, 1.8561260693675472)

    def test_third_method_single_value(self):
        result = tp.conversion.actual_vapour_pressure(tmin=TMIN, tmax=TMAX, rh_min=RHMIN, rh_max=RHMAX)
        np.testing.assert_almost_equal(result, 1.4058885969470276)

    def test_fourth_method_single_value(self):
        result = tp.conversion.actual_vapour_pressure(tmin=TMIN, rh_max=RHMAX)
        np.testing.assert_almost_equal(result, 1.1995003886672477)

    def test_fifth_method_single_value(self):
        result = tp.conversion.actual_vapour_pressure(rh_mean=RHMEAN, tmin=TMIN, tmax=TMAX)
        np.testing.assert_almost_equal(result, 1.4652762230908918)

    def test_sixth_method_single_value(self):
        result = tp.conversion.actual_vapour_pressure(tmin=TMIN)
        np.testing.assert_almost_equal(result, 1.4279766531752949)

    def test_raise_value_error(self):
        with self.assertRaises(ValueError):
            tp.conversion.actual_vapour_pressure(tmax=TMAX)


class TestMonthlySoilHeatFlux(unittest.TestCase):

    tmean_month_1 = 14.1
    tmean_month_2 = 16.1
    tmean_month_3 = 18.8

    def test_next_month_single_value(self):
        result = tp.conversion.monthly_soil_heat_flux(self.tmean_month_1, self.tmean_month_3, next_month=True)
        np.testing.assert_almost_equal(result, 0.33, decimal=1)

    def test_current_month_single_value(self):
        result = tp.conversion.monthly_soil_heat_flux(self.tmean_month_1, self.tmean_month_2)
        np.testing.assert_almost_equal(result, 0.28, decimal=1)


class testVectorsToScalar(unittest.TestCase):
    time = pd.date_range("2021-01-01", "2021-01-31")
    lat = range(0, 6)
    lon = range(0, 6)
    u10 = np.ones((len(lat), len(lon), time.size)) * np.arange(time.size)
    v10 = np.ones((len(lat), len(lon), time.size)) * np.arange(time.size)
    ds = xr.Dataset(
        data_vars=dict(
            u10=(["lat", "lon", "time"], u10),
            v10=(["lat", "lon", "time"], v10),
        ),
        coords=dict(lat=lat, lon=lon, time=time),
    )

    def test_value(self):
        exp = np.ones((len(self.lat), len(self.lon), self.time.size)) * np.sqrt(
            2 * np.arange(self.time.size) * np.arange(self.time.size)
        )

        obs = tp.conversion.vectors_to_scalar(self.ds["u10"], self.ds["v10"])
        np.testing.assert_almost_equal(exp, obs)


class testWindDirection(unittest.TestCase):

    time = pd.date_range("2021-01-01", "2021-01-16")
    lat = range(0, 6)
    lon = range(0, 6)
    tan30 = 1 / np.sqrt(3)
    tan60 = np.sqrt(3)
    ones = np.ones((len(lat), len(lon), time.size))
    u10 = ones * np.array([1, 0, -1, 0, 1, 1, -1, -1, 1, 1, tan60, tan30, -1, -1, -tan60, -tan30])
    v10 = ones * np.array([0, 1, 0, -1, 1, -1, -1, 1, -tan60, -tan30, 1, 1, tan60, tan30, -1, -1])
    ds = xr.Dataset(
        data_vars=dict(
            u10=(["lat", "lon", "time"], u10),
            v10=(["lat", "lon", "time"], v10),
        ),
        coords=dict(lat=lat, lon=lon, time=time),
    )

    def test_values(self):
        expected = self.ones * np.array(
            [270.0, 180.0, 90.0, 0.0, 225.0, 315.0, 45.0, 135.0, 330.0, 300.0, 240.0, 210.0, 150.0, 120.0, 60.0, 30.0]
        )

        result = tp.conversion.wind_direction(u=self.ds["u10"], v=self.ds["v10"], convention="from", unit="deg")
        np.testing.assert_almost_equal(expected, result)

    def test_unit_rads(self):
        exp = self.ones * np.array(
            [270.0, 180.0, 90.0, 0.0, 225.0, 315.0, 45.0, 135.0, 330.0, 300.0, 240.0, 210.0, 150.0, 120.0, 60.0, 30.0]
        )
        exp = exp * np.pi / 180

        obs = tp.conversion.wind_direction(u=self.ds["u10"], v=self.ds["v10"], convention="from", unit="rad")
        np.testing.assert_almost_equal(exp, obs)

    def test_direction_to(self):
        exp = self.ones * np.array(
            [270.0, 180.0, 90.0, 0.0, 225.0, 315.0, 45.0, 135.0, 330.0, 300.0, 240.0, 210.0, 150.0, 120.0, 60.0, 30.0]
        )
        exp = exp - 180
        exp = np.where(exp >= 0, exp, exp + 360)

        obs = tp.conversion.wind_direction(u=self.ds["u10"], v=self.ds["v10"], convention="to", unit="deg")
        np.testing.assert_almost_equal(exp, obs)

    def test_zero_wind(self):
        obs = tp.conversion.wind_direction(u=0, v=0)
        self.assertAlmostEqual(0, obs)

    def test_invalid_unit(self):
        with self.assertRaises(ValueError):
            tp.conversion.wind_direction(u=self.ds["u10"], v=self.ds["v10"], convention="from", unit="apple")

    def test_invalid_convention(self):
        with self.assertRaises(ValueError):
            tp.conversion.wind_direction(u=self.ds["u10"], v=self.ds["v10"], convention="oranges", unit="deg")


class TestRelativeHumidityFromActualAndSaturatedVapourPressure(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.relative_humidity_from_actual_vapour_pressure_and_saturated_vapour_pressure(AVP, SVP)
        np.testing.assert_almost_equal(result, 70.55583375062594)


class TestRelativeHumidityFromDewPointAndTemperature(unittest.TestCase):
    def test_single_value(self):
        result = tp.conversion.relative_humidity_from_dew_point_and_temperature(TDEW, T)
        np.testing.assert_almost_equal(result, 87.44015847592662)



import datetime as dt
import unittest
import zoneinfo
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from freezegun import freeze_time

import terrapyn as tp

PACKAGE_ROOT_DIR = Path(__file__).resolve().parent.parent
TEST_DATA_PATH = PACKAGE_ROOT_DIR / "tests" / "data"

idx = pd.IndexSlice


class TestConvertDatetime64(unittest.TestCase):
    def test_object_type(self):
        result = tp.time.datetime64_to_datetime(np.datetime64("2013-04-05 07:12:34.056789"))
        self.assertEqual(result, dt.datetime(2013, 4, 5, 7, 12, 34, 56789))


class TestConvertDatetime(unittest.TestCase):
    def test_object_type(self):
        result = tp.time.datetime_to_datetime64(dt.datetime(2013, 4, 5, 7, 12, 34, 123))
        self.assertEqual(result, np.datetime64("2013-04-05 07:12:34.000123"))


class TestGetTimeFromData(unittest.TestCase):
    expected = pd.DatetimeIndex(
        ["2019-03-15", "2019-03-16", "2019-03-17"], dtype="datetime64[ns]", name="time", freq=None
    )
    df = pd.DataFrame(
        {
            "time": expected,
            "id": [123, 456, 789],
            "val": [1, 3, 5],
        }
    ).set_index(["time", "id"])

    def test_dataframe(self):
        results = tp.time.get_time_from_data(self.df.reset_index(drop=False))
        pd.testing.assert_index_equal(results, self.expected)

    def test_dataframe_time_column(self):
        results = tp.time.get_time_from_data(self.df)
        pd.testing.assert_index_equal(results, self.expected)

    def test_dataset(self):
        results = tp.time.get_time_from_data(self.df.to_xarray())
        pd.testing.assert_index_equal(results, self.expected)

    def test_list(self):
        results = tp.time.get_time_from_data(list(self.expected.to_pydatetime()))
        pd.testing.assert_index_equal(results, self.expected)

    def test_dataarray(self):
        results = tp.time.get_time_from_data(self.df.to_xarray()["val"])
        pd.testing.assert_index_equal(results, self.expected)

    def test_series_time_index(self):
        results = tp.time.get_time_from_data(self.df["val"])
        pd.testing.assert_index_equal(results, self.expected)

    def test_series_time_column(self):
        results = tp.time.get_time_from_data(pd.Series(self.expected))
        pd.testing.assert_index_equal(results, self.expected)

    def test_datetime(self):
        results = tp.time.get_time_from_data(dt.datetime(2019, 3, 15))
        pd.testing.assert_index_equal(results, pd.DatetimeIndex([dt.datetime(2019, 3, 15)], name="time"))

    def test_ndarray(self):
        results = tp.time.get_time_from_data(self.expected.to_numpy())
        pd.testing.assert_index_equal(results, self.expected)

    def test_datetimeindex(self):
        results = tp.time.get_time_from_data(self.expected)
        pd.testing.assert_index_equal(results, self.expected)

    def test_multiindex(self):
        results = tp.time.get_time_from_data(self.df.index)
        pd.testing.assert_index_equal(results, self.expected)

    def test_missing_from_multiindex(self):
        with self.assertRaises(ValueError):
            tp.time.get_time_from_data(self.df.index, time_dim="date")

    def test_invalid_datatype(self):
        with self.assertRaises(TypeError):
            tp.time.get_time_from_data(1)


class TestGroupbyTime(unittest.TestCase):

    ds = xr.open_dataset(TEST_DATA_PATH / "lat_2_lon_2_time_15_D_test_data.nc")

    def test_dataset_groupby_week(self):
        result = tp.time.groupby_time(self.ds, grouping="week")
        self.assertEqual(result.groups, {8: [0, 1, 2, 3, 4], 9: [5, 6, 7, 8, 9, 10, 11], 10: [12, 13, 14]})

    def test_dataarray_groupby_week(self):
        result = tp.time.groupby_time(self.ds["var"], grouping="week")
        self.assertEqual(result.groups, {8: [0, 1, 2, 3, 4], 9: [5, 6, 7, 8, 9, 10, 11], 10: [12, 13, 14]})

    def test_dataset_groupby_month(self):
        result = tp.time.groupby_time(self.ds, grouping="month")
        self.assertEqual(result.groups, {2: [0, 1, 2, 3, 4, 5, 6, 7, 8], 3: [9, 10, 11, 12, 13, 14]})

    def test_dataframe_groupby_pentad(self):
        result = tp.time.groupby_time(self.ds.to_dataframe(), grouping="pentad")
        np.testing.assert_almost_equal(result.sum().values, np.array([[272.59223017], [257.44398154], [295.72954042]]))

    def test_series_groupby_dekad(self):
        result = tp.time.groupby_time(self.ds.to_dataframe()["var"], grouping="dekad")
        np.testing.assert_almost_equal(result.sum().values, np.array([80.23334597, 412.85991647, 332.67248968]))

    def test_groupby_year(self):
        result = tp.time.groupby_time(self.ds, grouping="year")
        self.assertEqual(result.groups, {2019: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]})

    def test_groupby_dayofyear(self):
        result = tp.time.groupby_time(self.ds, grouping="dayofyear")
        self.assertEqual(result.groups[55], [4])

    def test_invalid_grouping(self):
        with self.assertRaises(ValueError):
            tp.time.groupby_time(self.ds, grouping="10d")

    def test_groupby_other_keys(self):
        result = tp.time.groupby_time(self.ds.to_dataframe(), grouping="year", other_grouping_keys="lon").groups[
            (2019, 30)
        ][0]
        self.assertEqual(result, (pd.Timestamp("2019-02-20 00:00:00"), 2, 30))

    def test_groupby_multiple_other_keys(self):
        result = tp.time.groupby_time(
            self.ds.to_dataframe(), grouping="year", other_grouping_keys=["lat", "lon"]
        ).groups[(2019, 2, 30)][0]
        self.assertEqual(result, (pd.Timestamp("2019-02-20 00:00:00"), 2, 30))


@freeze_time("2012-01-14 11:12:13")
class TestDailyDateRange(unittest.TestCase):
    def test_date_today(self):
        result = tp.time.daily_date_range()
        self.assertEqual(result, [dt.datetime(2012, 1, 14, 0, 0)])

    def test_delta_days(self):
        result = tp.time.daily_date_range(delta_days=-1)
        self.assertEqual(result, [dt.datetime(2012, 1, 13, 0, 0), dt.datetime(2012, 1, 14, 0, 0)])

    def test_delta_days_future(self):
        result = tp.time.daily_date_range(delta_days=1)
        self.assertEqual(result, [dt.datetime(2012, 1, 14, 0, 0), dt.datetime(2012, 1, 15, 0, 0)])

    def test_future_dates(self):
        result = tp.time.daily_date_range(end_time=dt.datetime(2012, 1, 16), reset_time=True, hours=6)
        self.assertEqual(
            result, [dt.datetime(2012, 1, 14, 6, 0), dt.datetime(2012, 1, 15, 6, 0), dt.datetime(2012, 1, 16, 6, 0)]
        )


@freeze_time("2012-01-14 11:12:13")
class TestMonthlyDateRange(unittest.TestCase):
    def test_date_today_ignore_days(self):
        result = tp.time.monthly_date_range()
        self.assertEqual(result, [dt.datetime(2012, 1, 1, 0, 0)])

    def test_date_today_include_days(self):
        result = tp.time.monthly_date_range(reset_time=False)
        self.assertEqual(result, [dt.datetime(2012, 1, 14, 11, 12, 13)])

    def test_zero_delta_months(self):
        result = tp.time.monthly_date_range(delta_months=0)
        self.assertEqual(result, [dt.datetime(2012, 1, 1, 0, 0)])

    def test_delta_months(self):
        result = tp.time.monthly_date_range(delta_months=-2)
        self.assertEqual(
            result,
            [
                dt.datetime(2011, 11, 1, 0, 0),
                dt.datetime(2011, 12, 1, 0, 0),
                dt.datetime(2012, 1, 1, 0, 0),
            ],
        )


class TestAddDayOfYearVariable(unittest.TestCase):

    ds = xr.open_dataset(TEST_DATA_PATH / "lat_2_lon_2_time_366_D_test_data.nc")

    def test_dataset_modify_ordinal_days(self):
        result = tp.time.add_day_of_year_variable(self.ds)
        np.testing.assert_equal(result["dayofyear"].values[58:62], np.array([59, 60, 60, 61]))

    def test_dataset_no_modify_ordinal_days(self):
        result = tp.time.add_day_of_year_variable(self.ds, modify_ordinal_days=False)
        np.testing.assert_equal(result["dayofyear"].values[58:62], np.array([59, 60, 61, 62]))

    def test_dataarray(self):
        result = tp.time.add_day_of_year_variable(self.ds["var"], modify_ordinal_days=False)
        np.testing.assert_equal(result["dayofyear"].values[58:62], np.array([59, 60, 61, 62]))


class TestCheckStartEndTimeValidity(unittest.TestCase):
    def test_datetime64_with_datetime64(self):
        result = tp.time.check_start_end_time_validity(
            np.datetime64("2013-04-05 07:12:34.056789"), np.datetime64("2013-04-05 07:12:35")
        )
        self.assertTrue(result)

    def test_datetime64_with_datetime64_invalid(self):
        result = tp.time.check_start_end_time_validity(
            np.datetime64("2013-04-05 07:12:35"), np.datetime64("2013-04-05 07:12:34.056789")
        )
        self.assertFalse(result)

    def test_datetime_with_datetime64(self):
        result = tp.time.check_start_end_time_validity(
            dt.datetime(2013, 4, 5, 7, 12, 34, 56789), np.datetime64("2013-04-05 07:12:35")
        )
        self.assertTrue(result)

    @pytest.fixture(autouse=True)
    def capfd(self, capfd):
        self.capfd = capfd

    def test_verbose_warning(self):
        result = tp.time.check_start_end_time_validity(dt.datetime(2014, 1, 2), dt.datetime(2014, 1, 1), verbose=True)
        out, err = self.capfd.readouterr()
        self.assertFalse(result)
        assert out == "Warning: End time 2014-01-01 00:00:00 before start time 2014-01-02 00:00:00\n"

    def test_missing_time(self):
        with self.assertRaises(ValueError):
            tp.time.check_start_end_time_validity(None, dt.datetime(2014, 1, 2))


class TestGetDayOfYear(unittest.TestCase):

    ds = xr.open_dataset(TEST_DATA_PATH / "lat_2_lon_2_time_366_D_test_data.nc")

    def test_dataset_ordinal_days(self):
        result = tp.time.get_day_of_year(self.ds, modify_ordinal_days=False)[58:62]
        np.testing.assert_equal(result, np.array([59, 60, 61, 62]))

    def test_dataset_modify_ordinal_days(self):
        result = tp.time.get_day_of_year(self.ds, modify_ordinal_days=True)[58:62]
        np.testing.assert_equal(result, np.array([59, 60, 60, 61]))

    def test_dataarray_ordinal_days(self):
        result = tp.time.get_day_of_year(self.ds["var"], modify_ordinal_days=False)[58:62]
        np.testing.assert_equal(result, np.array([59, 60, 61, 62]))

    def test_dataframe_ordinal_days(self):
        result = tp.time.get_day_of_year(self.ds.to_dataframe(), time_dim="time", modify_ordinal_days=False)[58:62]
        np.testing.assert_equal(result, np.array([15, 15, 16, 16]))

    def test_series_ordinal_days(self):
        result = tp.time.get_day_of_year(self.ds.to_dataframe()["var"].index, modify_ordinal_days=False)[58:62]
        np.testing.assert_equal(result, np.array([15, 15, 16, 16]))

    def test_datetime_ordinal_days(self):
        result = tp.time.get_day_of_year(dt.datetime(2004, 3, 1), modify_ordinal_days=False)
        np.testing.assert_equal(result, np.array([61]))


class TestTimeToLocalTime(unittest.TestCase):
    def test_invalid_string(self):
        with self.assertRaises(TypeError):
            tp.time.time_to_local_time(dt.datetime(2019, 3, 15, 1, 0), timezone_name=1)


class TestDataToLocalTime(unittest.TestCase):
    expected = pd.DatetimeIndex(
        ["2019-03-15 01:00:00", "2019-03-15 02:00:00"], dtype="datetime64[ns]", name="time", freq=None
    )
    df = pd.DataFrame(
        {"time": pd.date_range("2019-03-15", freq="h", periods=2), "id": ["a", "b"], "val": [1, 2]}
    ).set_index(["time", "id"])

    def test_dataframe_multiindex(self):
        results = tp.time.data_to_local_time(self.df.copy(), "CET").index.get_level_values("time")
        pd.testing.assert_index_equal(results, self.expected)

    def test_dataframe(self):
        test_df = self.df.copy().reset_index(drop=False).set_index("time")
        results = tp.time.data_to_local_time(test_df, "CET").index.get_level_values("time")
        pd.testing.assert_index_equal(results, self.expected)

    def test_dataframe_column(self):
        results = tp.time.data_to_local_time(self.df.reset_index(drop=False).copy(), "CET", time_dim="time")["time"]
        np.testing.assert_equal(results.values, self.expected.values)

    def test_series(self):
        results = tp.time.data_to_local_time(self.df.copy()["val"], "CET").index.get_level_values("time")
        pd.testing.assert_index_equal(results, self.expected)

    def test_series_values(self):
        results = tp.time.data_to_local_time(pd.Series(self.expected), "CET").values
        np.testing.assert_equal(results, pd.Series(self.expected) + pd.Timedelta("1h"))

    def test_dataset(self):
        results = tp.time.data_to_local_time(self.df.copy().to_xarray(), "CET").indexes["time"]
        pd.testing.assert_index_equal(results, self.expected)

    def test_dataarray(self):
        results = tp.time.data_to_local_time(self.df.copy().to_xarray(), "CET")["val"].indexes["time"]
        pd.testing.assert_index_equal(results, self.expected)

    def test_datetime(self):
        results = tp.time.data_to_local_time(dt.datetime(2019, 3, 15, 1, 0), "CET")[0]
        self.assertTrue(results, dt.datetime(2019, 3, 15, 2, 0))

    def test_missing_timezone(self):
        with self.assertRaises(ValueError):
            tp.time.data_to_local_time(dt.datetime(2019, 3, 15, 1, 0), None)

    def test_datetimeindex(self):
        results = tp.time.data_to_local_time(self.expected - pd.Timedelta("1h"), "CET")
        pd.testing.assert_index_equal(results, self.expected)

    def test_ndarray(self):
        results = tp.time.data_to_local_time(self.expected.values - np.timedelta64(1, "h"), "CET")
        pd.testing.assert_index_equal(results, self.expected)

    def test_list(self):
        results = tp.time.data_to_local_time([dt.datetime(2019, 3, 15, 1, 0)], "CET")
        self.assertEqual(results.to_pydatetime(), [dt.datetime(2019, 3, 15, 2, 0)])

    @pytest.fixture(autouse=True)
    def capfd(self, capfd):
        self.capfd = capfd

    def test_invalid_data_type(self):
        with self.assertRaises(TypeError):
            out, err = self.capfd.readouterr()
            _ = tp.time.data_to_local_time(1, "CET")
            assert out == "Data type of int not implemented"


class TestListTimezones(unittest.TestCase):
    def test_dict_type(self):
        result = tp.time.list_timezones()
        self.assertTrue(isinstance(result, set))


class TestEnsureDatetimeIndex(unittest.TestCase):
    def test_datetime(self):
        result = tp.time._ensure_datetimeindex(dt.datetime(2021, 4, 5))
        expected = pd.DatetimeIndex(["2021-04-05 00:00:00"], dtype="datetime64[ns]", name="time", freq=None)
        pd.testing.assert_index_equal(result, expected)

    def test_list_of_datetimes(self):
        result = tp.time._ensure_datetimeindex([dt.datetime(2021, 4, 5), dt.datetime(2021, 4, 6)])
        expected = pd.DatetimeIndex(
            ["2021-04-05 00:00:00", "2021-04-06 00:00:00"], dtype="datetime64[ns]", name="time", freq=None
        )
        pd.testing.assert_index_equal(result, expected)

    def test_datetimeindex(self):
        expected = pd.DatetimeIndex(["2021-04-05 00:00:00"], dtype="datetime64[ns]", name="time", freq=None)
        result = tp.time._ensure_datetimeindex(expected)
        pd.testing.assert_index_equal(result, expected)


class TestDatetimeToUTC(unittest.TestCase):
    def test_no_timezone(self):
        result = tp.time._datetime_to_UTC(dt.datetime(2021, 4, 5))
        expected = pd.DatetimeIndex(["2021-04-05 00:00:00+00:00"], dtype="datetime64[ns, UTC]", name="time", freq=None)
        pd.testing.assert_index_equal(result, expected)

    def test_timezone_set(self):
        result = tp.time._datetime_to_UTC(dt.datetime(2021, 4, 5, tzinfo=zoneinfo.ZoneInfo("CET")))
        expected = pd.DatetimeIndex(["2021-04-04 22:00:00+00:00"], dtype="datetime64[ns, UTC]", name="time", freq=None)
        pd.testing.assert_index_equal(result, expected)


class TestDatetimeindexToLocalTimeTzAware(unittest.TestCase):
    def test_no_timezone(self):
        result = tp.time._datetimeindex_to_local_time_tz_aware(dt.datetime(2021, 4, 5))
        expected = pd.DatetimeIndex(["2021-04-05 00:00:00+00:00"], dtype="datetime64[ns, UTC]", name="time", freq=None)
        pd.testing.assert_index_equal(result, expected)

    def test_timezone_set(self):
        result = tp.time._datetimeindex_to_local_time_tz_aware(
            dt.datetime(2021, 4, 5, tzinfo=zoneinfo.ZoneInfo("CET")), "EST"
        )
        expected = pd.DatetimeIndex(["2021-04-04 17:00:00-05:00"], dtype="datetime64[ns, EST]", name="time", freq=None)
        pd.testing.assert_index_equal(result, expected)


class TestDatetimeindexToLocalTimeTzNaive(unittest.TestCase):
    def test_no_timezone(self):
        result = tp.time._datetimeindex_to_local_time_tz_naive(dt.datetime(2021, 4, 5))
        expected = pd.DatetimeIndex(["2021-04-05"], dtype="datetime64[ns]", name="time", freq=None)
        pd.testing.assert_index_equal(result, expected)

    def test_timezone_set(self):
        result = tp.time._datetimeindex_to_local_time_tz_naive(dt.datetime(2021, 4, 5, tzinfo=zoneinfo.ZoneInfo("CET")))
        expected = pd.DatetimeIndex(["2021-04-04 22:00:00"], dtype="datetime64[ns]", name="time", freq=None)
        pd.testing.assert_index_equal(result, expected)


class TestSetTimeInData(unittest.TestCase):

    df = pd.DataFrame(
        {"time": pd.date_range("2019-03-15 06:00", freq="D", periods=2), "id": ["a", "b"], "val": [1, 2]}
    ).set_index(["time", "id"])

    def test_replace_times(self):
        results = tp.time._set_time_in_data(
            self.df, new_times=pd.date_range("2021-01-1 06:00", freq="D", periods=2)
        ).index.get_level_values("time")
        expected = pd.DatetimeIndex(
            ["2021-01-01 06:00:00", "2021-01-02 06:00:00"], dtype="datetime64[ns]", name="time", freq=None
        )
        pd.testing.assert_index_equal(results, expected)

    def test_set_times_to_midnight(self):
        results = tp.time._set_time_in_data(
            self.df, set_time_to_midnight=True, hours_to_subtract=None
        ).index.get_level_values("time")
        expected = pd.DatetimeIndex(["2021-01-01", "2021-01-02"], dtype="datetime64[ns]", name="time", freq=None)
        pd.testing.assert_index_equal(results, expected)

    def test_subtract_hours(self):
        results = tp.time._set_time_in_data(
            self.df, set_time_to_midnight=True, hours_to_subtract=5
        ).index.get_level_values("time")
        expected = pd.DatetimeIndex(
            ["2020-12-31 19:00:00", "2021-01-01 19:00:00"], dtype="datetime64[ns]", name="time", freq=None
        )
        pd.testing.assert_index_equal(results, expected)

    def test_no_modification(self):
        results = tp.time._set_time_in_data(self.df).index.get_level_values("time")
        expected = self.df.index.get_level_values("time")
        pd.testing.assert_index_equal(results, expected)


class TestUTCOffsetInHours(unittest.TestCase):
    def test_datetimeindex_no_timezone(self):
        result = tp.time.utc_offset_in_hours(pd.date_range("2019-01-02", freq="6H", periods=2), "Asia/Kolkata")
        self.assertEqual(result, 5.5)

    def test_datetime_no_timezone(self):
        result = tp.time.utc_offset_in_hours(dt.datetime(2021, 4, 5), "Asia/Kolkata")
        self.assertEqual(result, 5.5)

    def test_datetimeindex_timezone_set(self):
        result = tp.time.utc_offset_in_hours(
            pd.date_range("2019-01-02", freq="6H", periods=2, tz="CET"), "Asia/Kolkata"
        )
        self.assertEqual(result, 5.5)

    def test_datetimeindex_return_multiple_offsets(self):
        result = tp.time.utc_offset_in_hours(
            pd.date_range("2019-01-02", freq="6H", periods=2), "Asia/Kolkata", return_single_value=False
        )
        self.assertEqual(result, [5.5, 5.5])


class TestGroupbyFreq(unittest.TestCase):
    ds = xr.Dataset(
        data_vars={"var": (("lat", "lon", "time"), np.ones((1, 1, 100)))},
        coords={"lat": [1], "lon": [2], "time": pd.date_range("2022-01-01", periods=100)},
    )

    ds_hourly = xr.Dataset(
        data_vars={"var": (("lat", "lon", "time"), np.arange(100)[np.newaxis, np.newaxis, :])},
        coords={"lat": [1], "lon": [2], "time": pd.date_range("2022-01-01", periods=100, freq="h")},
    )

    ds_hourly_multicoord = xr.Dataset(
        data_vars={"var": (("lat", "lon", "time"), np.full((2, 2, 100), np.arange(100)))},
        coords={"lat": [1, 2], "lon": [3, 4], "time": pd.date_range("2022-01-01", periods=100, freq="h")},
    )

    def test_dataset(self):
        result = tp.time.groupby_freq(self.ds, freq="M").sum()["var"].values.flatten()
        expected = np.array([30.0, 28.0, 31.0, 11.0])
        np.testing.assert_array_equal(result, expected)

    def test_dataarray(self):
        result = tp.time.groupby_freq(self.ds["var"], freq="M").sum().values.flatten()
        expected = np.array([30.0, 28.0, 31.0, 11.0])
        np.testing.assert_array_equal(result, expected)

    def test_dataframe(self):
        result = tp.time.groupby_freq(self.ds["var"].to_dataframe(), freq="M").sum()["var"].values
        expected = np.array([30.0, 28.0, 31.0, 11.0])
        np.testing.assert_array_equal(result, expected)

    def test_series(self):
        result = tp.time.groupby_freq(self.ds["var"].to_series(), freq="M").sum().values
        expected = np.array([30.0, 28.0, 31.0, 11.0])
        np.testing.assert_array_equal(result, expected)

    def test_dataset_hourly_to_daily(self):
        result = tp.time.groupby_freq(self.ds_hourly, freq="D", day_start_hour=6).sum()["var"].values.flatten()
        expected = np.array([15, 420, 996, 1572, 1947])
        np.testing.assert_array_equal(result, expected)

    def test_dataframe_hourly_to_daily(self):
        result = tp.time.groupby_freq(self.ds_hourly.to_dataframe(), freq="D", day_start_hour=6).sum()["var"]
        expected = np.array([15, 420, 996, 1572, 1947])
        np.testing.assert_array_equal(result.values, expected)
        self.assertEqual(result.index.names, ["time", "lat", "lon"])

    def test_series_hourly_to_daily_single_index(self):
        result = tp.time.groupby_freq(
            self.ds_hourly["var"].to_series().reset_index(drop=True, level=["lat", "lon"]), freq="D", day_start_hour=6
        ).sum()
        expected = np.array([15, 420, 996, 1572, 1947])
        np.testing.assert_array_equal(result.values, expected)
        self.assertEqual(result.index.name, "time")

    def test_dataframe_no_time_dim(self):
        with self.assertRaises(ValueError):
            tp.time.groupby_freq(self.ds["var"].to_dataframe(), freq="M", time_dim="x")

    def test_series_no_time_dim(self):
        with self.assertRaises(ValueError):
            tp.time.groupby_freq(self.ds["var"].to_series(), freq="M", time_dim="x")

    def test_dataframe_time_column(self):
        result = (
            tp.time.groupby_freq(self.ds["var"].to_dataframe().reset_index(drop=False), freq="M").sum()["var"].values
        )
        expected = np.array([30.0, 28.0, 31.0, 11.0])
        np.testing.assert_array_equal(result, expected)

    def test_dataframe_time_column_other_cols(self):
        result = (
            tp.time.groupby_freq(
                self.ds_hourly_multicoord.to_dataframe().reset_index(drop=False), freq="D", other_grouping_columns="lat"
            )
            .sum()["lon"]
            .values
        )
        expected = np.array([168, 168, 168, 168, 168, 168, 168, 168, 28, 28])
        np.testing.assert_array_equal(result, expected)

    def test_invalid_datatype(self):
        with self.assertRaises(TypeError):
            tp.time.groupby_freq([1])


class TestResampleTime(unittest.TestCase):
    ds_hourly = xr.Dataset(
        data_vars={"var": (("lat", "lon", "time"), np.arange(100)[np.newaxis, np.newaxis, :])},
        coords={"lat": [1], "lon": [2], "time": pd.date_range("2022-01-01", periods=100, freq="h")},
    )

    def test_dataset_sum(self):
        result = tp.time.resample_time(self.ds_hourly, freq="D", day_start_hour=6, resample_method="sum")
        expected = np.array([15, 420, 996, 1572, 1947])
        self.assertEqual(result["time"].values[0], np.datetime64("2021-12-31T06:00:00.00"))
        np.testing.assert_array_equal(result["var"].values.flatten(), expected)

    def test_dataset_mean(self):
        result = tp.time.resample_time(self.ds_hourly, resample_method="mean")["var"].values.flatten()
        np.testing.assert_array_equal(result, np.array([11.5, 35.5, 59.5, 83.5, 97.5]))

    def test_dataset_max(self):
        result = tp.time.resample_time(self.ds_hourly, resample_method="max")["var"].values.flatten()
        np.testing.assert_array_equal(result, np.array([23, 47, 71, 95, 99]))

    def test_dataset_min(self):
        result = tp.time.resample_time(self.ds_hourly, resample_method="min")["var"].values.flatten()
        np.testing.assert_array_equal(result, np.array([0, 24, 48, 72, 96]))

    def test_method_not_implemented(self):
        with self.assertRaises(ValueError):
            tp.time.resample_time(self.ds_hourly, resample_method="foobar")



import numpy as np

from terrapyn import conversion
from terrapyn.validation import check_day_hours as _check_day_hours
from terrapyn.validation import check_day_of_year as _check_day_of_year
from terrapyn.validation import check_latitude_radians as _check_latitude_radians
from terrapyn.validation import (
    check_solar_declination_radians as _check_solar_declination_radians,
)
from terrapyn.validation import (
    check_sunset_hour_angle_radians as _check_sunset_hour_angle_radians,
)

# Solar constant
SOLAR_CONSTANT = 117.57312  # [ MJ m-2 day-1] = 1360.8 W/m2

# Stefan Boltzmann constant
STEFAN_BOLTZMANN_CONSTANT = 0.000000004899203  # [MJ K-4 m-2 day-1]


def _solar_parameters(latitude=None, day_of_year=None, sunshine_hours=None, altitude=None, coastal=False):
    """
    Calculate and return solar declination, sunset hour angle, inverse earth-sun distance,
    and extraterrestrial radiation (helper function, as these parameters are commonly used together).

    Args:
        latitude: Latitude [radians]
        day_of_year: Day of year, between 1 and 366.
        sunshine_hours: Total sunshine duration for that latitude, for that day [hours].
        altitude: Altitude/elevation of point above sea level [m]
        coastal: ``True`` if site is a coastal location, situated on or adjacent to coast of a large land mass
        and where air masses are influenced by a nearby water body, ``False`` if interior location where land
        mass dominates and air masses are not strongly influenced by a large water body.

    """
    if latitude is None:
        raise ValueError("'latitude' is not given")
    if day_of_year is None:
        raise ValueError("'day_of_year' is not given")

    sol_dec = solar_declination(day_of_year)
    sha = sunset_hour_angle(latitude, sol_dec)
    ird = inverse_relative_distance_earth_sun(day_of_year)
    et_rad = extraterrestrial_radiation(latitude, sol_dec, sha, ird)
    return (
        sol_dec,
        sha,
        ird,
        et_rad,
    )


def clear_sky_radiation(altitude=None, et_rad=None):
    """
    Estimate clear sky radiation from altitude and extraterrestrial radiation.
    Based on equation 37 in Allen et al (1998) which is recommended when calibrated Angstrom values
    are not available.

    Args:
        altitude: Elevation above sea level [m]
        et_rad: Extraterrestrial radiation [MJ m-2 day-1]. Can be estimated using ``extraterrestrial_radiation()``.

    Returns:
        Clear sky radiation [MJ m-2 day-1]
    """
    if altitude is None:
        raise ValueError("`altitude` must be provided")
    if et_rad is None:
        raise ValueError("`et_rad` must be provided")
    return (0.00002 * altitude + 0.75) * et_rad


def total_incoming_shortwave_radiation_from_sunshine_hours(dl_hours=None, sunshine_hours=None, et_rad=None):
    """
    Calculate incoming shortwave radiation `Rs` (radiation hitting a horizontal plane
    after scattering by the atmosphere) from relative sunshine duration. If measured radiation data
    are not available, this method is preferable to calculating solar radiation from temperature.
    If a monthly mean is required, then divide the monthly number of sunshine hours by number of
    days in the month and ensure that `et_rad` and `dl_hours` was calculated using the day of the
    year that corresponds to the middle of the month.
    Based on equations 34 and 35 in Allen et al (1998).

    Args:
        dl_hours: Number of daylight hours [hours]. Can be calculated using `daylight_hours()`.
        sunshine_hours: Sunshine duration [hours]. If not available, use
        `total_incoming_shortwave_radiation_from_temperature()` instead of this function.
        et_rad: Extraterrestrial radiation [MJ m-2 day-1]. Can be estimated using `extraterrestrial_radiation()`.

    Returns:
        Incoming solar (or shortwave) radiation [MJ m-2 day-1]
    """
    for variable, name in zip([dl_hours, sunshine_hours, et_rad], ["dl_hours", "sunshine_hours", "et_rad"]):
        if variable is None:
            raise ValueError("`{}` must be provided".format(name))

    _check_day_hours(sunshine_hours)
    _check_day_hours(dl_hours)

    # 0.5 and 0.25 are default values of regression constants (Angstrom values)
    # recommended by FAO when calibrated values are unavailable.
    return (0.5 * sunshine_hours / dl_hours + 0.25) * et_rad


def total_incoming_shortwave_radiation_from_temperature(
    et_rad=None, cs_rad=None, altitude=None, tmin=None, tmax=None, coastal=False
):
    """
    Estimate incoming shortwave radiation `Rs` (radiation hitting a horizontal plane after scattering by the
    atmosphere) from min and max temperature together with an empirical adjustment coefficient for 'interior'
    and 'coastal' regions. The formula is based on equation 50 in Allen et al (1998) which is the
    Hargreaves radiation formula (Hargreaves and Samani, 1982, 1985). This method should be used
    only when solar radiation or sunshine hours data are not available. It is only recommended for locations where
    it is not possible to use radiation data from a regional station (either because climate conditions are
    heterogeneous or data are lacking).
    **NOTE**: this method is not suitable for island locations due to the moderating effects of the surrounding water.

    Args:
        et_rad: Extraterrestrial radiation [MJ m-2 day-1]. Can be estimated using `extraterrestrial_radiation()`.
        cs_rad: Clear sky radiation [MJ m-2 day-1]. If not available, can provide `altitude` instead.
        altitude: Altitude/elevation of point above sea level [m]. Only required if `cs_rad` is not given.
        tmin: Daily minimum temperature [deg C].
        tmax: Daily maximum temperature [deg C].
        coastal: ``True`` if site is a coastal location, situated on or adjacent to coast of a large land mass
        and where air masses are influenced by a nearby water body, ``False`` if interior location where land
        mass dominates and air masses are not strongly influenced by a large water body.

    Returns:
        Incoming solar (or shortwave) radiation (Rs) [MJ m-2 day-1].
    """
    for variable, name in zip([tmax, tmin, et_rad], ["tmax", "tmin", "et_rad"]):
        if variable is None:
            raise ValueError("`{}` must be provided".format(name))

    if cs_rad is None:
        cs_rad = clear_sky_radiation(altitude, et_rad)

    # Determine value of adjustment coefficient [deg C-0.5] for coastal/interior locations
    if coastal:
        adj = 0.19
    else:
        adj = 0.16

    total_incoming_shortwave_rad = adj * np.sqrt(tmax - tmin) * et_rad

    # The solar radiation value is constrained by the clear sky radiation
    return np.minimum(total_incoming_shortwave_rad, cs_rad)


def total_incoming_shortwave_radiation(
    dl_hours=None,
    sunshine_hours=None,
    et_rad=None,
    cs_rad=None,
    altitude=None,
    tmin=None,
    tmax=None,
    coastal=False,
):
    """
    Calculate the total incoming shortwave radiation per day `Rs` (radiation hitting a horizontal plane after
    scattering by the atmosphere) in units of [MJ m-2 day-1].

    This is a wrapper that automatically chooses between two other functions:
    `total_incoming_shortwave_radiation_from_sunshine_hours()` or
    `total_incoming_shortwave_radiation_from_temperature()`.

    The method of estimation is chosen automatically depending on which parameters are given, where
    the preferred option is to provide `dl_hours`, `sunshine_hours`, and `et_rad`. Otherwise, `et_rad`,
    `cs_rad`/`altitude`, `tmin`, `tmax`, and `coastal` should be provided.

    If a monthly mean is required, divide the monthly number of sunshine hours by number of days in the
    month and ensure that `et_rad` and `dl_hours` are calculated using the day of the year that
    corresponds to the middle of the month.

    Args:
        dl_hours: Number of daylight hours [hours]. Can be calculated using `daylight_hours()`.
        sunshine_hours: Sunshine duration [hours].
        altitude: Altitude/elevation of point above sea level [m]
        et_rad: Extraterrestrial radiation [MJ m-2 day-1]. Can be estimated using `extraterrestrial_radiation()`.
        cs_rad: Clear sky radiation [MJ m-2 day-1]. Can be estimated using `clear_sky_radiation()`.
        tmin: Daily minimum temperature [deg C].
        tmax: Daily maximum temperature [deg C].
        coastal: ``True`` if site is a coastal location, situated on or adjacent to coast of a large land mass
        and where air masses are influenced by a nearby water body, ``False`` if interior location where land
        mass dominates and air masses are not strongly influenced by a large water body.

    Returns:
        Incoming solar (or shortwave) radiation (Rs) [MJ m-2 day-1].
    """
    # If either of df_hours and sunshine_hours are not provided, estimate total incoming shortwave
    # radiation from temperature
    if None in [dl_hours, sunshine_hours]:
        total_incoming_shortwave_rad = total_incoming_shortwave_radiation_from_temperature(
            et_rad=et_rad, cs_rad=cs_rad, altitude=altitude, tmin=tmin, tmax=tmax, coastal=coastal
        )

    # Else estimate total incoming shortwave radiation from sunshine hours and daylight hours
    else:
        total_incoming_shortwave_rad = total_incoming_shortwave_radiation_from_sunshine_hours(
            dl_hours, sunshine_hours, et_rad
        )

    return total_incoming_shortwave_rad


def net_incoming_shortwave_radiation(
    total_incoming_shortwave_rad=None,
    albedo=0.23,
    sunshine_hours=None,
    dl_hours=None,
    et_rad=None,
    cs_rad=None,
    altitude=None,
    tmin=None,
    tmax=None,
    coastal=False,
):
    """
    Calculate net incoming shortwave radiation from gross (total) incoming shortwave solar radiation,
    and the albedo (reflectance) from the surface. Net incoming shortwave radiation is equal to the incoming shortwave
    radiation minus the radiation lost from reflected solar radiation. The output can be converted to
    equivalent evaporation [mm day-1] using `energy_to_evaporation()`.

    The default value of `albedo=0.23` assumes a grass reference crop. Based on FAO equation 38 in Allen et al (1998).

    Args:
        total_incoming_shortwave_rad: Total incoming shortwave radiation [MJ m-2 day-1]. If given, no other arguments
        are required, apart from `albedo`. If not provided, this is estimated by passing all other arguments of
        this function (excluding `albedo`) to `total_incoming_shortwave_radiation()`.
        albedo: Albedo of the crop as the proportion of gross incoming solar radiation that is reflected by the
        surface. Default value is 0.23, which is the value used by the FAO for a short grass reference crop. Albedo
        can be as high as 0.95 for freshly fallen snow and as low as 0.05 for wet bare soil. A green vegetation over
        has an albedo of about 0.20-0.25 (Allen et al, 1998).
        latitude: Latitude [radians] (Not required if `total_incoming_shortwave_rad` is provided).
        day_of_year: Day of year, between 1 and 366. (Not required if `total_incoming_shortwave_rad` is provided).
        sunshine_hours: Total sunshine duration for that latitude, for that day [hours]. (Not required if
        `total_incoming_shortwave_rad` is provided).
        altitude: Altitude/elevation of point above sea level [m]. (Not required if `total_incoming_shortwave_rad`
        is provided).
        coastal: ``True`` if site is a coastal location, situated on or adjacent to coast of a large land mass
        and where air masses are influenced by a nearby water body, ``False`` if interior location where land
        mass dominates and air masses are not strongly influenced by a large water body. (Not required if
        `total_incoming_shortwave_rad` is provided).

    Returns:
        Net incoming shortwave radiation [MJ m-2 day-1] (net incoming minus reflected).
    """
    if total_incoming_shortwave_rad is None:
        total_incoming_shortwave_rad = total_incoming_shortwave_radiation(
            dl_hours=dl_hours,
            sunshine_hours=sunshine_hours,
            et_rad=et_rad,
            cs_rad=cs_rad,
            altitude=altitude,
            tmin=tmin,
            tmax=tmax,
            coastal=coastal,
        )
    return (1 - albedo) * total_incoming_shortwave_rad


def net_outward_longwave_radiation(
    tmin=None,
    tmax=None,
    total_incoming_shortwave_rad=None,
    et_rad=None,
    cs_rad=None,
    avp=None,
    latitude=None,
    day_of_year=None,
    dl_hours=None,
    sunshine_hours=None,
    altitude=None,
    coastal=False,
    tdew=None,
    twet=None,
    tdry=None,
    rh_min=None,
    rh_max=None,
    atmos_pres=None,
    rh_mean=None,
):
    """
    Estimate net outgoing longwave radiation. This is the net longwave energy (net energy flux) leaving the
    earth's surface. It is proportional to the absolute temperature of the surface raised to the fourth power
    according to the Stefan-Boltzmann law. However, water vapour, clouds, carbon dioxide and dust are absorbers
    and emitters of longwave radiation. This function corrects the Stefan-Boltzmann law for humidity
    (using actual vapor pressure) and cloudiness (using solar radiation and clear sky radiation). The concentrations
    of all other absorbers are assumed to be constant.
    The output can be converted to equivalent evaporation [mm day-1] using ``energy_to_evaporation()``.
    Based on FAO equation 39 in Allen et al (1998).

    Args:
        tmin: Daily minimum temperature [deg C]
        tmax: Daily maximum temperature [deg C]
        total_incoming_shortwave_rad: Total incoming shortwave radiation [MJ m-2 day-1]. Not required
        if `latitude`, `day_of_year`, `sunshine_hours`, `altitude`, `coastal` are provided.
        et_rad: Extraterrestrial radiation [MJ m-2 day-1]. Can be estimated using `extraterrestrial_radiation()`.
        cs_rad: Clear sky radiation [MJ m-2 day-1]. Optional, calculated automatically with other parameters.
        avp: Actual vapour pressure [kPa]. Can be estimated from multiple methods, in order of preference:

            1. If dewpoint temperature data are available, based on `tdew`.
            2. If dry and wet bulb temperatures are available from a psychrometer,
            based on `twet`, `tdry`, `atmos_pres`.
            3. If  minimum and maximum relative humidity data available, based on `tmin`,
            `tmax`, `rh_min`, `rh_max`.
            4. If measurement errors of relative humidity are large then use only maximum
            relative humidity, based on `tmin`, `rh_max`.
            5. If minimum and maximum relative humidity are not available but mean relative
            humidity is available, based on `tmin`, `tmax`, `rh_mean` (less reliable than options 3 or 4).
            6. If no data for the above, based on `tmin`. This function is less reliable in arid areas
            where it is recommended that 2 degrees Celsius is subtracted from the minimum temperature
            before it is passed to the function (following advice given in Annex 6 of Allen et al (1998).

        latitude: Latitude [radians] - should be negative if it in the southern hemisphere, positive if in
        the northern hemisphere.
        day_of_year: Day of year integer between 1 and 366.
        dl_hours: Number of daylight hours [hours]. Can be calculated using `daylight_hours()`.
        sunshine_hours: Sunshine duration [hours].
        altitude: Altitude/elevation of point above sea level [m]
        coastal: ``True`` if site is a coastal location, situated on or adjacent to coast of a large land mass
        and where air masses are influenced by a nearby water body, ``False`` if interior location where land
        mass dominates and air masses are not strongly influenced by a large water body.
        tdew: Dew point temperature [deg C].
        twet: Wet buld temperature [deg C].
        tdry: Dry air temperature [deg C].
        rh_min: Minimum relative humidity [%].
        rh_max: Maximum relative humidity [%].
        atmos_pres: Atmospheric pressure [kPa].
        rh_mean: Mean relative humidity [kPa].

    Return:
        Net outgoing longwave radiation [MJ m-2 day-1]
    """
    if tmax is None:
        raise ValueError("'tmax' must be provided")
    if tmin is None:
        raise ValueError("'tmin' must be provided")

    if total_incoming_shortwave_rad is None:
        total_incoming_shortwave_rad = total_incoming_shortwave_radiation(
            dl_hours=dl_hours,
            sunshine_hours=sunshine_hours,
            et_rad=et_rad,
            cs_rad=cs_rad,
            altitude=altitude,
            tmin=tmin,
            tmax=tmax,
            coastal=coastal,
        )
    if cs_rad is None:
        sol_dec, sha, ird, et_rad = _solar_parameters(
            latitude=latitude,
            day_of_year=day_of_year,
            sunshine_hours=sunshine_hours,
            altitude=altitude,
            coastal=coastal,
        )
        cs_rad = clear_sky_radiation(altitude=altitude, et_rad=et_rad)

    if avp is None:
        avp = conversion.actual_vapour_pressure(
            tdew=tdew,
            twet=twet,
            tdry=tdry,
            atmos_pres=atmos_pres,
            tmin=tmin,
            tmax=tmax,
            rh_min=rh_min,
            rh_max=rh_max,
            rh_mean=rh_mean,
        )

    tmp1 = (np.power(conversion.celsius_to_kelvin(tmax), 4) + np.power(conversion.celsius_to_kelvin(tmin), 4)) / 2
    tmp2 = 0.34 - (0.14 * np.sqrt(avp))
    tmp3 = 1.35 * (total_incoming_shortwave_rad / cs_rad) - 0.35
    return STEFAN_BOLTZMANN_CONSTANT * tmp1 * tmp2 * tmp3


def daylight_hours(sha=None, latitude=None, sol_dec=None, day_of_year=None):
    """
    Calculate daylight hours from sunset hour angle. Optionally, use latitude, and solar declination or day of year.
    Based on FAO equation 34 in Allen et al (1998).

    Args:
        sha: Sunset hour angle [rad]. Not required if `latitude` and `sol_dec`, or `day_of_year` are provided.
        latitude: Latitude [radians] -  should be negative if it in the southern hemisphere, positive if in
        the northern hemisphere. Not required if `sha` is provided.
        sol_dec: Solar declination [radians].  Not required if `day_of_year` or `sha` is provided.
        day_of_year: Day of year integer between 1 and 365 or 366. Not required if `sol_dec` or `sha` is provided.

    Returns:
        Daylight hours [h].
    """
    if sha is None:
        if latitude is None:
            raise ValueError("Either `sha` or `latitude` must be provided")
        else:
            sha = sunset_hour_angle(latitude=latitude, sol_dec=sol_dec, day_of_year=day_of_year)
    else:
        _check_sunset_hour_angle_radians(sha)
    return (24.0 / np.pi) * sha


def sunset_hour_angle(latitude=None, sol_dec=None, day_of_year=None):
    """
    Calculate sunset hour angle (*Ws*) from latitude and solar declination.
    Based on FAO equation 25 in Allen et al (1998).

    Args:
        latitude: Latitude [radians] -  should be negative if it in the southern hemisphere,
        positive if in the northern hemisphere.
        sol_dec: Solar declination [radians].  Not required if `day_of_year` is provided.

    Returns:
        Sunset hour angle [radians].
    """
    _check_latitude_radians(latitude)

    if sol_dec is None:
        if day_of_year is None:
            raise ValueError("Either `sol_dec` or `day_of_year` must be provided")
        else:
            sol_dec = solar_declination(day_of_year)
    else:
        _check_solar_declination_radians(sol_dec)

    cos_sha = -np.tan(latitude) * np.tan(sol_dec)

    # Domain of arccos is -1 <= x <= 1 radians (this is not mentioned in FAO-56)
    # See https://itaca.stuffnting.com/the-sun-as-a-source-of-energy/part-3-calculating-solar-angles/
    # Ensure cos_sha values lie within the range -1 <= cos_sha <= 1

    # If cos_sha is >= 1 there is no sunset, i.e. 24 hours of daylight
    cos_sha_no_sunset = np.where(cos_sha > 1.0, 1, cos_sha)

    # If cos_sha is <= 1 there is no sunrise, i.e. 24 hours of darkness
    cos_sha_no_sunrise_sunset = np.where(cos_sha_no_sunset < -1.0, -1, cos_sha_no_sunset)
    return np.arccos(cos_sha_no_sunrise_sunset)


def solar_declination(day_of_year=None):
    """
    Calculate solar declination from day of the year.
    Based on FAO equation 24 in Allen et al (1998).

    Args:
        day_of_year: Day of year integer between 1 and 365 or 366.

    Returns:
        solar declination [radians]
    """
    _check_day_of_year(day_of_year)
    return 0.409 * np.sin(((2.0 * np.pi / 365.0) * day_of_year - 1.39))


def extraterrestrial_radiation(latitude=None, sol_dec=None, sha=None, ird=None, day_of_year=None):
    #                          solar_time_start=None, solar_time_end=None
    """
    Estimate daily extraterrestrial radiation (*Ra*, 'top of the atmosphere radiation').
    Based on equation 21 in Allen et al (1998). If monthly mean radiation is required make sure
    *sol_dec*, *sha* and *irl* have been calculated using the day of the year that corresponds to
    the middle of the month. **Note**: From Allen et al (1998): "For the winter months in latitudes
    greater than 55 degrees (N or S), the equations have limited validity. Reference should be made
    to the Smithsonian Tables to assess possible deviations."

    Args:
        latitude: Latitude [radians] - should be negative if in the southern hemisphere, positive
        if in the northern hemisphere.
        sol_dec: Solar declination [radians]. Not required if `day_of_year` is provided.
        sha: Sunset hour angle [radians]. Not required if `latitude` and `sol_dec` or `day_of_year` are provided.
        ird: Inverse relative distance earth-sun [dimensionless]. Not required if `day_of_year` is provided.
        day_of_year: Day of year (1-365/366). Not required if `ird` and `sol_dec` are provided.

    TODO:
        Add ability to calculate radiation for periods less than 1 day - see eqn 28 from Allen
        solar_time_start: Start solar time for periods less than 1 day.
        solar_time_end: End solar time for periods less than 1 day.

    Returns:
        Daily extraterrestrial radiation [MJ m-2 day-1]
    """
    _check_latitude_radians(latitude)

    if ird is None:
        if day_of_year is None:
            raise ValueError("Either `ird` or `day_of_year` must be provided")
        else:
            ird = inverse_relative_distance_earth_sun(day_of_year)

    if sol_dec is None:
        if day_of_year is None:
            raise ValueError("Either `sol_dec` or `day_of_year` must be provided")
        else:
            sol_dec = solar_declination(day_of_year)
    else:
        _check_solar_declination_radians(sol_dec)

    if sha is None:
        sha = sunset_hour_angle(latitude, sol_dec)
    else:
        _check_sunset_hour_angle_radians(sha)

    tmp1 = SOLAR_CONSTANT / np.pi
    tmp2 = sha * np.sin(latitude) * np.sin(sol_dec)
    tmp3 = np.cos(latitude) * np.cos(sol_dec) * np.sin(sha)
    return tmp1 * ird * (tmp2 + tmp3)


def inverse_relative_distance_earth_sun(day_of_year=None):
    """
    Calculate the inverse relative distance between earth and sun from day of the year.
    Based on FAO equation 23 in Allen et al (1998).

    Args:
        day_of_year: Day of the year [1 to 366]

    Returns:
        Inverse relative distance between earth and the sun
    """
    _check_day_of_year(day_of_year)
    return 1 + (0.033 * np.cos((2.0 * np.pi / 365.0) * day_of_year))



import os

import setuptools

version = os.getenv("VERSION", "0.0.9")

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

with open("dev-requirements.txt") as f:
    dev_install_requires = f.read().splitlines()

setuptools.setup(
    name="terrapyn",
    version=version,
    url="https://github.com/colinahill/terrapyn",
    description="Toolkit to manipulate Earth observations and models.",
    author="Colin Hill",
    author_email="colinalastairhill@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    readme="README.md",
    license="BSD-3-Clause",
    install_requires=install_requires,
    packages=setuptools.find_packages(exclude=("site",)),
    extras_require={"dev": dev_install_requires},
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.10",
    ],
    include=("LICENSE", "terrapyn/py.typed"),
)



import typing as T

import numpy as np
import pandas as pd


def nearest_multiple(x, base: float = 0.5) -> np.ndarray:
    """
    Round a value to the nearest multiple of a given base. Rounding follows `numpy.round`,
    so 0.5 is closer to 1 than 0. Works for single values or arrays/lists.

    Args:
        x: Value/array
        base: Base used to round the value.

    Returns:
        Rounded value(s)

    :Example:
    >>> nearest_multiple(6.73, base=0.02)
    6.72
    >>> nearest_multiple([1.77, 1.771], base=0.02)
    array([1.76, 1.78])
    """
    x = np.float64(x)
    base = np.float64(base)
    return np.round(np.divide(x, base), decimals=0) * base


def sigma_clip(
    data: T.Union[np.ndarray, pd.Series] = None,
    low_sigma: T.Union[int, float] = 7,
    upp_sigma: T.Union[int, float] = 7,
    n_iter: int = None,
    return_flags: bool = True,
    return_thresholds: bool = False,
) -> T.Union[
    T.Union[np.ndarray, pd.Series], T.Tuple[T.Union[np.ndarray, pd.Series], T.Optional[float], T.Optional[float]]
]:
    r"""
    Calculates the mean and standard deviation (`std`) of an array, and detemines if the values lie outside
    the range :math:`(mean - low_sigma * std)` to :math:`(mean + upp_sigma * std)`. `NaN` values are always
    flagged as bad points. If `return_flags==True`, returns an array mask where `True` indicates the value
    is rejected/bad (lying outside the defined range).

    If `n_iter` is `None`, perform iterative sigma-clipping of array elements. Starting from the full sample,
    all elements outside the critical range are removed, i.e. until all elements of `data` satisfy both conditions::

        data < mean(data) - std(data) * low_sigma
        data > mean(data) + std(data) * upp_sigma

    The iteration continues with the updated sample until no elements are outside the (updated) range.
    If `n_iter` is given, the iterative process is repeated `n_iter` times.

    Args:
        data: Input data
        low_sigma: Number of standard deviations to use as the lower rejection threshold
        upp_sigma: Number of standard deviations to use as the upper rejection threshold
        n_iter: If `None`, perform iterative sigma-clipping of array elements until no elements
        are outside the (updated) critical range. Otherwise, if `n_iter` is given, repeate the
        iterative process `n_iter` times.
        return_flags: If `True`, return a boolean array mask where `True` indicates the value is
        outside the defined range. If `False`, return the input array with clipped elements removed
        return_thresholds: If `True`, return the lower and upper threshold values used for clipping.

    Returns:
        If `return_flags==True`, an array mask for the values, where `True` indicates the value is outside
        the defined range. If `return_flags==False`, the input array with clipped elements removed.
        if `return_thresholds==True`, additionally return the lower and upper threshold values used for clipping.
    """
    if n_iter is not None and n_iter < 1:
        raise ValueError("`n_iter` must be `None` or `>=1`")

    # Ensure data is a 1-D array
    if isinstance(data, pd.Series):
        array = data.values
    elif isinstance(data, np.ndarray):
        array = np.asarray(data).ravel()
    else:
        raise TypeError("`data` must by of type `pd.Series` or `np.ndarray`")

    if len(array.shape) > 1:
        raise ValueError("'data' must be 1-D")

    # Initialize starting mask, including only non-NaN points
    mask = np.isfinite(array)

    iteration = 0
    while True:
        iteration += 1

        # Mean and std of subset of elements, where `mask` are good points
        mean = array[mask].mean()
        std = array[mask].std()
        lower = mean - low_sigma * std
        upper = mean + upp_sigma * std

        # Size of subset
        size = array[mask].size

        # mask_temp is boolean mask for good points in whole array based on new rejection criteria
        mask_temp = (array >= lower) & (array <= upper)

        # Update mask with temporary mask
        mask = mask_temp

        # check to see if the size of the array is equal to the number of included elements. If True, nothing more to do
        # OR check if the required number of iterations has been reached
        if (size - np.sum(mask)) == 0 or (iteration == n_iter):
            break

    if return_flags:
        # `mask` is `True` for good points, so invert mask so `True` is a rejected point
        mask = np.logical_not(mask)
        if return_thresholds:
            return mask, lower, upper
        else:
            return mask
    else:
        if isinstance(data, pd.Series):
            array = data[mask]
        else:
            array = array[mask]

        if return_thresholds:
            return array, lower, upper
        else:
            return array



import typing as T

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr


def split_into_chunks(a: T.Iterable = None, n: int = 2, axis: int = 0) -> np.ndarray:
    """
    Split an array/list/iterable into chunks of length `n`, where the last chunk may have length < n.

    Args:
        a: Iterable array/list
        n: Length of each chunk
        axis: The axis along which to split the data
    Returns:
        Array split into chunks of length `n`

    Example:
        >>> split_into_chunks(np.arange(10), 3)
        [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8]), array([9])]
    """
    return np.split(a, range(n, len(a), n), axis=axis)


def set_dim_values_in_data(
    data: T.Union[
        xr.Dataset,
        xr.DataArray,
        pd.DataFrame,
        pd.Series,
    ] = None,
    values: T.Iterable = None,
    dim: str = None,
) -> T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series]:
    """
    Replaces the values of a dimension/variable/column/index named `dim` in
    xarray or pandas data structures with the values in the iterable `values`.

    Args:
        data: Input data with a column/dimension/index called `dim`.
        values: Values to use to replace the existing values for that index/dim/column.
        dim: Name of dimension/column/index (ignored for `pandas.Series`).

    Returns:
        Data with `dim` values replaced with `values`.
    """
    if isinstance(data, (pd.Series, pd.DataFrame)):
        if dim in data.index.names:
            if isinstance(data.index, pd.MultiIndex):
                data.index = data.index.set_levels(values, level=dim)
            else:
                data.index = values
        else:
            if isinstance(data, pd.Series):
                # Update values of series, not index
                data.update(values)
            else:
                # update column of dataframe
                data[dim] = values
    elif isinstance(data, (xr.Dataset, xr.DataArray)):
        data = data.assign_coords({dim: values})
    else:
        data_types_str = ", ".join(
            str(i)
            for i in [
                xr.Dataset,
                xr.DataArray,
                pd.DataFrame,
                pd.Series,
            ]
        )
        raise TypeError(f"Data must be one of type: {data_types_str}")
    return data


def get_dim_values_from_data(
    data: T.Union[
        xr.Dataset,
        xr.DataArray,
        pd.DataFrame,
        pd.Series,
        pd.MultiIndex,
    ] = None,
    dim: str = None,
) -> np.ndarray:
    """
    Retrieve the values of a dimension/variable/column/index in xarray or pandas data structures.
    If values are a pd.Index or pd.Series data type, these types are returned, otherwise a np.ndarray is returned.

    Args:
        data: Input data
        dim: Name of dimension/variable/column/index, the values of which will be returned

    Returns:
        Values for the dimension/variable/column/index `dim`
    """
    # Validate `dim` string
    if not isinstance(dim, str) or len(dim) == 0:
        raise ValueError("`dim` must be a string with length > 0")

    if isinstance(data, (pd.Series, pd.DataFrame)):
        # Check if `dim` is in the index
        if dim in data.index.names:
            return data.index.get_level_values(dim).values
        else:
            if isinstance(data, pd.Series):
                return data.values
            elif dim in data.columns:
                return data[dim].values
            else:
                raise ValueError(f"`dim` of '{dim} not found in data")
    elif isinstance(data, (xr.Dataset, xr.DataArray)):
        if dim in data.variables:
            return data.variables[dim].values
        else:
            raise ValueError(f"`dim` of '{dim} not found in data")
    elif isinstance(data, pd.MultiIndex):
        # Check if `dim` is the index
        if dim in data.names:
            return data.get_level_values(dim).values
        else:
            raise ValueError(f"`dim` of {dim} not found in pd.Multindex")
    else:
        data_types_str = ", ".join(str(i) for i in [xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, pd.Multindex])
        raise TypeError(f"Data is of type {type(data)} but must be one of type: {data_types_str}")


def pandas_to_geopandas(
    df: pd.DataFrame, lat_col: str = "lat", lon_col: str = "lon", crs=None, *args, **kwargs
) -> gpd.GeoDataFrame:
    """
    Convert a pandas.DataFrame to a geopandas.GeoDataFrame, adding a new geometry column
    with GeometryArray of shapely Point geometries.

    Args:
        df: Input Pandas dataframe with columns of latitude and longitude.
        lat_col: Name of column with latitudes
        lon_col: Name of column with longitudes
        crs: (Optional) Coordinate Reference System of the geometry objects

    Returns:
        Geopandas GeoDataFrame
    """
    gdf = gpd.GeoDataFrame(df, *args, **kwargs)
    gdf["geometry"] = gpd.points_from_xy(gdf[lon_col], gdf[lat_col], crs=crs)
    return gdf


def _pandas_check_multiindex_type(index):
    """Checks type of index and returns True if pd.Multiindex"""
    return isinstance(index, pd.MultiIndex)


def _dim_in_pandas_index(index, dim):
    """Check if dimension `dim` is in a Pandas index (of any type)"""
    is_multiindex = _pandas_check_multiindex_type(index)

    if is_multiindex:
        if index.names is not None:
            return dim in index.names
    else:
        if index.name is not None:
            return dim in index.name


def ensure_list(a: T.Any = None) -> T.List[T.Any]:
    """Ensure data `a` is a list and cast if required"""
    if isinstance(a, list):
        return a
    elif isinstance(a, (str, int, float)):
        return [a]
    else:
        return list(a)


def _call_resample_method(obj, method, **kwargs):
    """Apply resample method to grouped object, either pandas"""
    if method == "sum":
        return obj.sum(**kwargs)
    elif method == "mean":
        return obj.mean(**kwargs)
    elif method == "min":
        return obj.min(**kwargs)
    elif method == "max":
        return obj.max(**kwargs)
    # elif method == "cumsum":
    else:
        raise ValueError(f"method=`{method}` not implemented")


def remove_list_elements(input_list: T.List = None, remove_list: T.List = None) -> T.List:
    """Removes all elements in `remove_list` from `input_list` and returns new list object"""
    if input_list is not None and remove_list is not None:

        # convert all single values to a list
        remove_list = ensure_list(remove_list)

        return [e for e in input_list if e not in set(remove_list)]

    else:
        raise ValueError("both `input_list` and `remove_list` must be provided")
